/* Specification of the echo algorithm in Alloy 6
   using graphs with a dedicated initiating node
   author: Burkhardt Renz, November 2021
*/
 
open util/graph[Node]

let unchanged[r] { (r)' = (r) } 

enum Color {Red, Green}

abstract sig Node {
	neighbors: set Node,
	var parent: lone Node,
	var inbox: set Node,
	var color: Color
}	

one sig INode extends Node{}
sig PNode extends Node{}

fact {
	noSelfLoops[neighbors]
	undirected[neighbors]
	rootedAt[neighbors, INode]
}

pred init {
	parent = INode->INode
	no inbox
	color = Node->Red
}

pred stutter {
	unchanged[parent]
	unchanged[inbox]
	unchanged[color]
}

pred broadcast[n, fp: Node] {
	all q: n.neighbors - fp | q.inbox' = q.inbox + n
	all u: Node - n.neighbors + fp | u.inbox' = u.inbox	
}

pred initiate {
		init
		broadcast[INode, INode]
		unchanged[parent]
		unchanged[color]
}

pred forward [n: Node, msg: Node] {
	no n.parent and msg in n.inbox
	parent' = parent + n->msg
	broadcast[n, msg]
	unchanged[color]
}

pred echo [n: Node] {
	one n.parent and n.inbox = n.neighbors and n.color = Red
	unchanged[parent]
	n = INode implies
		inbox' = inbox
	else
		inbox' = inbox ++ n.parent->(n.parent.inbox + n)
		color' = color ++ (n->Green)
}

fact trans {
	init
	always { 
		initiate or
		some n, msg: Node | forward[n, msg] or
		some n: Node | echo[n] or
		stutter }
}

pred fairness {
	all n, msg: Node {
		(eventually always init) 
				implies (always eventually initiate)
		(eventually always (no n.parent and msg in n.inbox))
				implies (always eventually forward[n, msg])
		(eventually always (one n.parent and n.inbox = n.neighbors and n.color = Red))
				implies (always eventually echo[n])
	}
}
		
run examples {}

run withProgress {fairness} for exactly 4 Node

assert EchoComesBack {
	fairness implies eventually INode.color = Green
}

check EchoComesBack for 5 Node, 1.. steps

assert SpanningTree {
	let pt = parent - INode->INode |
		always INode.color = Green 
			implies (tree[~pt] and parent.Node = Node)
}

check SpanningTree //for 5 Node, 1.. steps

