/* Specification of the echo algorithm in Alloy 6
   using graphs without dedicated initiating node
   author: Burkhardt Renz, Nocember 2021
*/
module echo_var

open util/graph[Node]

let unchanged[r] { (r)' = (r) } 

enum Color {Red, Green}

sig Node {
	neighbors: set Node,
	var parent: lone Node,
	var inbox: set Node,
	var color: Color
}	

fact {
	noSelfLoops[neighbors]
	undirected[neighbors]
	stronglyConnected[neighbors]
}

pred stutter {
	unchanged[parent]
	unchanged[inbox]
	unchanged[color]
}

pred broadcast[n, fp: Node] {
	all q: n.neighbors - fp | q.inbox' = q.inbox + n
	all u: Node - n.neighbors + fp | u.inbox' = u.inbox	
}

pred initiate[n: Node] {
		parent = n -> n
		no inbox
		color = Node -> Red
	  unchanged[parent]
		#Node = 1 implies {
			color' = n -> Green 
			unchanged[inbox]
		} else {
			broadcast[n, n]
			unchanged[color]
	}
}

pred forward [n: Node, msg: Node] {
	no n.parent and msg in n.inbox
	parent' = parent + n->msg
	broadcast[n, msg]
	unchanged[color]
}

pred echo [n: Node] {
	one n.parent and n.inbox = n.neighbors and n.color = Red
	unchanged[parent]
	n.parent = n implies
		unchanged[inbox]
	else
		inbox' = inbox ++ n.parent->(n.parent.inbox + n)
	color' = color ++ (n->Green)
}

fact trans {
	some n: Node | initiate[n]
	after always { 
		some n, msg: Node | forward[n, msg] or
		some n: Node | echo[n] or
		stutter }
}

pred fairness {
	all n, msg: Node {
		(eventually always (no n.parent and msg in n.inbox))
				implies (always eventually forward[n, msg])
		(eventually always (one n.parent and n.inbox = n.neighbors and n.color = Red))
				implies (always eventually echo[n])
	}
}
								
run examples {}

run makeProgress {fairness} for exactly 5 Node, 1.. steps

assert echoComesBack {
	fairness implies (eventually all n:Node | n.color = Green)
}

check echoComesBack for 5 Node, 1.. steps

assert SpanningTree {
	always (color = Node->Green
		implies tree[~(parent-iden)])
}

check SpanningTree //for 5 Node, 1..steps

